#include "list.h"

static caddr_t Malloc(int _size)
{
  caddr_t pointer;

  pointer = (caddr_t)malloc(_size);

  if (pointer == (caddr_t)NULL)
  {
    printf("Malloc: out of memory.\n");
    exit(0);
  }

  return pointer;
}

void ListError(int _type)
{
  switch(_type)
  {
    case NO_COMPARE_FUNC:
      printf("Your compareFunc is NULL!\n");
      break;
    case LIST_HEAD_NULL:
      printf("The list is uninitialized.\n");
      break;
    case LIST_EMPTY:
      printf("The list is empty.\n");
      break;
    case DATA_EMPTY:
      printf("The data is empty!\n");
      break;
    case NOT_FOUND:
      printf("That piece of data is not in the list.\n");
      break;
    case NO_PRINT_FUNC:
      printf("No print function has been specified.\n");
      break;
    default:
      printf("Unknown error %d\n", _type);
      exit(0);
  }
}

List ListCreate(int (*compareFunc)(), void (*printFunc)())
{
  List l;

  if (compareFunc == NULL)
  {
    ListError(NO_COMPARE_FUNC);
    exit(0);
  }

  l = (List)Malloc(sizeof(ListHead));

  l->length = 0;
  l->list = NULL;
  l->compareFunc = compareFunc;
#if DEBUGGING
  l->printFunc = printFunc;
#endif /* DEBUGGING */

  return l;
}

int ListAdd(List *_l, caddr_t _data, int _where)
{
  ListPtr lp;

  if (*_l == (List)NULL)
  {
    ListError(LIST_HEAD_NULL);
    exit(0);
  }

  if (_data == NULL)
  {
    ListError(DATA_EMPTY);
    return ERROR;
  }

  lp = (ListPtr)Malloc(sizeof(ListRec));
  lp->data = _data;
  lp->next = NULL;
  (*_l)->length++;

  if (ListEmpty(*_l))
  {
    (*_l)->list = lp;
    return OK;
  }
  else
  if (_where == Head)
  {
    lp->next = (*_l)->list;
    (*_l)->list = lp;
    return OK;
  }
  else
  {
    ListPtr temp;

    if ((*_l)->list == (ListPtr)NULL)
    {
      (*_l)->list = lp;
      return OK;
    }

    temp = (*_l)->list;

    while(temp->next != (ListPtr)NULL)
      temp = temp->next;

    temp->next = lp;
    return OK;
  }
}

int ListRemove(List *_l, caddr_t _data)
{
  ListPtr temp, before;
  int Found = FALSE;
  int i;

  if (*_l == (List)NULL)
  {
    ListError(LIST_HEAD_NULL);
    exit(0);
  }

  if (ListEmpty(*_l))
    return LIST_EMPTY;

  if (_data == NULL)
  {
    ListError(DATA_EMPTY);
    return ERROR;
  }

  temp = (*_l)->list;
  Found = (*(*_l)->compareFunc)(_data, temp->data);

  if (Found)
    (*_l)->list = (*_l)->list->next;
  else
  {
    before = (*_l)->list;
    temp = before->next;
  }

  while((temp != NULL) && !Found)
  {
    Found = (*(*_l)->compareFunc)(_data, temp->data);
    if (!Found)
    {
      before = temp;
      temp = temp->next;
    }
    else 
      before->next = temp->next;
  }

  if (!Found)
    return NOT_FOUND;
 
  temp->data = NULL;
  free(temp);
  (*_l)->length--;
  return OK; 
}

int ListFind(List _l, caddr_t _data)
{
  int pos;
  ListPtr temp;
  int Found = FALSE;
  int i;

  if (_l == (List)NULL)
  {
    ListError(LIST_HEAD_NULL);
    exit(0);
  }

  if (_data == NULL)
  {
    ListError(DATA_EMPTY);
    return ERROR;
  }

  if (ListEmpty(_l))
    return NOT_FOUND;

  temp = _l->list;

  i = 0;
  while((temp != (ListPtr)NULL) && !Found)
  {
    i++;
    Found = _l->compareFunc(_data, temp->data);
    if (!Found) 
      temp = temp->next;
  }

  if (Found)
    return i;
  else
    return NOT_FOUND;
}

caddr_t ListFirst(List _l)
{
  if (_l == (List)NULL)
  {
    ListError(LIST_HEAD_NULL);
    exit(0);
  }

  if (_l->list == (ListPtr)NULL)
    return (caddr_t)NULL;

  return _l->list->data;
}

caddr_t ListLast(List _l)
{
  ListPtr temp;

  if (_l == (List)NULL)
  {
    ListError(LIST_HEAD_NULL);
    exit(0);
  }

  temp = _l->list;

  if (temp == (ListPtr)NULL)
    return (caddr_t)NULL;

  while(temp->next != (ListPtr)NULL)
    temp = temp->next;

  return temp->data;
}

int ListEmpty(List _l)
{
  if (_l == (List)NULL)
  {
    ListError(LIST_HEAD_NULL);
    exit(0);
  }

  return _l->list == (ListPtr)NULL;
}

int ListLength(List _l)
{
  if (_l == (List)NULL)
  {
    ListError(LIST_HEAD_NULL);
    exit(0);
  }

  return _l->length;
}

void ListDestroy(List *_l)
{
  ListPtr temp, temp2;

  if (*_l == (List)NULL)
    return;

  temp = (*_l)->list;
  while(temp != (ListPtr)NULL)
  {
    temp->data = NULL;
    temp2 = temp->next;
    free(temp);
    temp = temp2;
  }

  free(*_l);

  *_l = (List)NULL;
}

#if DEBUGGING
void PrintList(List _l)
{
  ListPtr temp;
  int i = 1;

  if (_l == (List)NULL)
  {
    ListError(LIST_HEAD_NULL);
    exit(0);
  }

  if (_l->printFunc == NULL)
  {
    ListError(NO_PRINT_FUNC);
    return;
  }
 
  temp = _l->list;
  while(temp != (ListPtr)NULL)
  {
    printf("Item %d: ", i);
    (*_l->printFunc)(temp->data);
    temp = temp->next;
    i++;
  }
}
#endif /* DEBUGGING */

